#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

#define CUDA_RT_CALL(call)                                                                  \
    {                                                                                       \
        cudaError_t cudaStatus = call;                                                      \
        if (cudaSuccess != cudaStatus)                                                      \
            fprintf(stderr,                                                                 \
                    "ERROR: CUDA RT call \"%s\" in line %d of file %s failed "              \
                    "with "                                                                 \
                    "%s (%d).\n",                                                           \
                    #call, __LINE__, __FILE__, cudaGetErrorString(cudaStatus), cudaStatus); \
    }
__global__ void hello(const char * payload){
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    
    if (i == 0) {
        printf("%s", payload);
    }
}

int main(int argc, char** argv){
    MPI_Init(&argc, &argv);

    int size;
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    MPI_Status stat;

    if(size != 2){
        if(rank == 0){
            printf("This program requires exactly 2 MPI ranks, but you are attempting to use %d! Exiting...\n", size);
        }
        MPI_Finalize();
        exit(0);
    }
    int tag = 10;

    const char *payload = "hello ";

    char * d_payload;
    CUDA_RT_CALL( cudaMalloc((void**)&d_payload, 6) );
    CUDA_RT_CALL( cudaMemcpy(d_payload, payload, 6, cudaMemcpyHostToDevice) );

    hello<<<1, 1>>>(d_payload);

    CUDA_RT_CALL( cudaPeekAtLastError() );
    CUDA_RT_CALL( cudaDeviceSynchronize() );
    
    // TODO: Implement the MPI_Recv() call to receive the "payload" from rank 1 using directly "d_payload" as the target buffer on the GPU
    if (rank == 1) {
    #ifdef SOLUTION
        MPI_Recv(d_payload, 6, MPI_CHAR, 0, tag, MPI_COMM_WORLD, &stat);
    #else
        MPI_Recv();
    #endif
    }
    
    hello<<<1, 1>>>(d_payload);

    CUDA_RT_CALL( cudaPeekAtLastError() );
    CUDA_RT_CALL( cudaDeviceSynchronize() );
    
    printf("\n");
    
    MPI_Finalize();
    return 0;
}
